/*
 * Copyright (c) 1991-2023 by STEP Tools Inc. 
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and
 * its documentation is hereby granted, provided that this copyright
 * notice and license appear on all copies of the software.
 * 
 * STEP TOOLS MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE
 * SUITABILITY OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING
 * BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE, OR NON-INFRINGEMENT. STEP TOOLS
 * SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY LICENSEE AS A
 * RESULT OF USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE OR ITS
 * DERIVATIVES.
 *
 * Author: David Loffredo (loffredo@steptools.com)
 */

#include <stp_schema.h>
#include "toolpath.h"

void ToolpathMaker::GoToXYZ(
    const char * label, 
    double x, 
    double y, 
    double z
    )
{
    if (!f_design) return;


    // if we were in 5-axis mode and the tool axis not is at 0,0,1,
    // then generate a 5-axis move to the new location and move the
    // tool axis to 0,0,1, otherwise start a 3-axis toolpath.
    //
    if (axis.last_point_id &&
	! (STEPNC_FLOAT_EQUAL (axis.last_point[0], 0) &&
	   STEPNC_FLOAT_EQUAL (axis.last_point[1], 0) &&
	   STEPNC_FLOAT_EQUAL (axis.last_point[2], 1)) )
    {
	GoToXYZ_IJK(label, x,y,z, 0,0,1);
	end_move(); // next move is just a 3-axis move
	axis.last_point_id = 0;
	return;
    }

    if (axis.last_point_id) {
	end_move(); // go from 5-axis to 3-axis toolpaths
	axis.last_point_id = 0;
    }
    
    start_move (label);

    // if this is the first point in a polyline and we are at a
    // particular location from a previous toolpath, reuse the last
    // point from the previous move as our starting point.
    //
    if (loc.last_point_id && !loc.points.size())
    {
	loc.points.append (loc.last_point_id);
    }

    loc.last_point_id = make_point (label, x, y, z);
    loc.last_point[0] = x;
    loc.last_point[1] = y;
    loc.last_point[2] = z;
    
    loc.points.append (loc.last_point_id);
}


void ToolpathMaker::GoToXYZ_IJK(
    const char * label, 
    double x, 
    double y, 
    double z,
    double i,
    double j, 
    double k
    )
{
    if (!f_design) return;

    start_move (label);

    // If our last move was in 3axis mode, close any existing toolpath
    // and start a new 5-axis one from the last point.
    //
    if (loc.last_point_id && !axis.last_point_id)
    {
	// If we have more than a single point, we must start a new
	// toolpath.  Add the last location point as our initial point
	// with a 0,0,1 tool axis vector.
	if (loc.curves.size() || loc.points.size() > 1) {
	    end_move();
	    start_move (label);
	}

	// A single initial 3-axis move is ok, othewise the toolpath
	// will be separated and we have to re-add the last point
	if (!loc.points.size()) 
	    loc.points.append (loc.last_point_id);

	// the default tool axis
	axis.points.append (make_point ("", 0,0,1));
    }

    // if this is the first point in a polyline and we are at a
    // particular location from a previous toolpath, reuse the last
    // point from the previous move as our starting point.
    //
    if (loc.last_point_id && !loc.points.size())
    {
	loc.points.append (loc.last_point_id);

	// if we were not in 5-axis mode, assume the tool axis is 0,0,1.
	//
	if (!axis.last_point_id) {
	    axis.last_point_id = make_point (label, 0,0,1);
	}
	axis.points.append (axis.last_point_id);
    }

    loc.last_point_id = make_point (label, x, y, z);
    loc.last_point[0] = x;
    loc.last_point[1] = y;
    loc.last_point[2] = z;
    loc.points.append (loc.last_point_id);


    // If the tool axis has not changed, we can reuse the old one,
    // otherwise we have to create a new one as below.
    //
    if (!axis.last_point_id ||
	! (STEPNC_FLOAT_EQUAL (axis.last_point[0], i) &&
	   STEPNC_FLOAT_EQUAL (axis.last_point[1], j) &&
	   STEPNC_FLOAT_EQUAL (axis.last_point[2], k)) )
    {
	axis.last_point_id = make_point (label, i, j, k);
	axis.last_point[0] = i;
	axis.last_point[1] = j;
	axis.last_point[2] = k;
    }
    axis.points.append (axis.last_point_id);
}


void ToolpathMaker::Arc(
    const char * label, 
    double new_x, 
    double new_y, 
    double new_z, 
    double cx, 
    double cy, 
    double cz, 
    double rad, 
    int ccw
    )
{
    if (!f_design) return;

    start_move (label);

    // No rapids in arcs because the machines do not do that.
    if (is_rapid) {
	fprintf (stderr, "%s : RAPID arc not allowed\n", 
		 label? label: "arc");
	return;
    }

    // Must have done at least one goto to give the start point
    if (!loc.last_point_id) {
	fprintf (stderr, "%s : Missing goto for start point of arc", 
		 label? label: "arc");
	return;
    }

    stp_cartesian_point * arc_start = loc.last_point_id;

    // arcs can only be described as part of a composite curve, so we
    // must start a new composite curve segment.  Close out a polyline
    // that happens to be open.  If the list only has one point then
    // we do not need to make a polyline.
    //
    if (loc.points.size() == 1) {
	loc.points.empty();
    }
    if (loc.points.size() > 1) {
	loc.curves.append ((stp_curve*) make_polyline ("", loc.points));
	loc.points.empty();
    }

    // Make a trimmed curve based upon a circle to represent the arc.
    // Add it to the composite curve segments so that we can write
    // everything out when we end the toolpath.
    RoseStringObject nm;
    nm.resize (strlen(label? label: "") + 100);
    sprintf (nm,"Arc end for %s in WS %u TP %u", 
	     (label? label: "-"), ws_count, tp_count);

    stp_cartesian_point * arc_end = make_point (nm, new_x,new_y,new_z);

    sprintf (nm,"Arc center for %s in WS %u TP %u", 
	     (label? label: "-"), ws_count, tp_count);

    stp_cartesian_point * arc_center = make_point (nm, cx,cy,cz);


    // Write an axis placement with the canonical Z and X directions
    if (!current_zdir || !current_xdir) 
    {
	SetDirection (0,0,1, 1,0,0);
    }

    sprintf (nm,"Arc placement for %s in WS %u TP %u", 
	     (label? label: "-"), ws_count, tp_count);

    stp_axis2_placement_3d * arc_ap3d = pnewIn(f_design) 
	stp_axis2_placement_3d;
    arc_ap3d-> name (nm);
    arc_ap3d-> location (arc_center);
    arc_ap3d-> axis (current_zdir);
    arc_ap3d-> ref_direction (current_xdir);

    // Write a circle at that location with the provided radius
    sprintf (nm,"Circle for %s in WS %u TP %u", 
	     (label? label: "-"), ws_count, tp_count);

    stp_circle * arc_circle = pnewIn(f_design) stp_circle;
    arc_circle-> name (nm);
    arc_circle-> radius (rad);
    arc_circle-> position (pnewIn(f_design) stp_axis2_placement);
    arc_circle-> position()-> _axis2_placement_3d (arc_ap3d);

    // Trim the circle at the start and end points
    sprintf (nm,"Arc for %s in WS %u TP %u", 
	     (label? label: "-"), ws_count, tp_count);

    stp_trimmed_curve * arc_trim = pnewIn(f_design) stp_trimmed_curve;
    arc_trim-> name (nm);
    arc_trim-> basis_curve (arc_circle);
    arc_trim-> sense_agreement (ccw? ROSE_TRUE: ROSE_FALSE);
    arc_trim-> master_representation(stp_trimming_preference_cartesian);
    arc_trim-> trim_1()-> add(pnewIn (f_design) stp_trimming_select);
    arc_trim-> trim_1()-> first()-> _cartesian_point (arc_start);
    arc_trim-> trim_2()-> add(pnewIn (f_design) stp_trimming_select);
    arc_trim-> trim_2()-> first()-> _cartesian_point (arc_end);
	
    // push the trimmed curve onto the composite curves segments list
    // so that we tie it together with the othe curves at the end of
    // the toolpath.
    loc.curves.append ((stp_curve*) arc_trim);
    loc.last_point_id = arc_end;
    loc.last_point[0] = new_x;
    loc.last_point[1] = new_y;
    loc.last_point[2] = new_z;

    
    // TOOL AXIS - if the arc is being machined in 5-axis mode, then
    // we need a two point poly line describing the tool axis at the
    // start and end.  These two points will be the same because cncs
    // can only do an arc in a plane.  It is not clear whether the
    // tool axis and the normal for the circle must be the same, but
    // we just use the incoming tool axis.
    //
    // As with the location curve, check to see if we have an open
    // polyline with a single point.  If we do, add one more point
    // otherwise we make a new polyline.
    //

    if (axis.last_point_id)
    {
	stp_cartesian_point * arc_axis = axis.last_point_id;

	// if we have a non-trivial open polyline, close it and start
	// another one
	//
	if (axis.points.size() > 1) {
	    axis.curves.append ((stp_curve*) make_polyline ("", axis.points));
	    axis.points.empty();
	}

	// create a two point polyline and append it.
	if (axis.points.size() == 1) {
	    axis.points.append (arc_axis);
	}
	else {
	    axis.points.append (arc_axis);
	    axis.points.append (arc_axis);
	}

	axis.curves.append ((stp_curve*) make_polyline ("", axis.points));
	axis.points.empty();
    }
}





void  ToolpathMaker::SetDirection(
    double zi, double zj, double zk, 	// z axis
    double xa, double xb, double xc		// x axis
    )
{
    if (!f_design) return;

    // We could check to make sure that Z dot X is not zero

    // Create directions with given Z and X directions.  Will be used
    // in the placement for the circle in ARCs.  Replace any previous
    // directions.
    //

    // Share the Z direction vector for 001
    if (STEPNC_FLOAT_EQUAL (zi, 0) && 
	STEPNC_FLOAT_EQUAL (zj, 0) &&
	STEPNC_FLOAT_EQUAL (zk, 1)) 
    {
	const char * shortcut = "__CONSTANT DIRECTION zaxis";

	/* Use shortcut to unit if we have already created one */
	RoseObject * obj = f_design-> findObject (shortcut);
	if (obj) 
	    current_zdir = ROSE_CAST (stp_direction, obj);
	else { 
	    current_zdir = pnewIn(f_design) stp_direction;
	    current_zdir-> name ("Z direction");
	    current_zdir-> direction_ratios()-> add (0);
	    current_zdir-> direction_ratios()-> add (0);
	    current_zdir-> direction_ratios()-> add (1);
	    f_design-> addName (shortcut, current_zdir);  // save shortcut
	}
    }
    else {
	current_zdir = pnewIn(f_design) stp_direction;
	current_zdir-> name ("Z direction");
	current_zdir-> direction_ratios()-> add (zi);
	current_zdir-> direction_ratios()-> add (zj);
	current_zdir-> direction_ratios()-> add (zk);
    }


    // Share the X direction vector for 100
    if (STEPNC_FLOAT_EQUAL (xa, 1) && 
	STEPNC_FLOAT_EQUAL (xb, 0) &&
	STEPNC_FLOAT_EQUAL (xc, 0)) 
    {
	const char * shortcut = "__CONSTANT DIRECTION xaxis";

	/* Use shortcut to unit if we have already created one */
	RoseObject * obj = f_design-> findObject (shortcut);
	if (obj) 
	    current_xdir = ROSE_CAST (stp_direction, obj);
	else { 
	    current_xdir = pnewIn(f_design) stp_direction;
	    current_xdir-> name ("Z direction");
	    current_xdir-> direction_ratios()-> add (1);
	    current_xdir-> direction_ratios()-> add (0);
	    current_xdir-> direction_ratios()-> add (0);
	    f_design-> addName (shortcut, current_xdir);  // save shortcut
	}
    }
    else {
	current_xdir = pnewIn(f_design) stp_direction;
	current_xdir-> name ("Z direction");
	current_xdir-> direction_ratios()-> add (xa);
	current_xdir-> direction_ratios()-> add (xb);
	current_xdir-> direction_ratios()-> add (xc);
    }
}



void ToolpathMaker::NestWorkplan (const char * name)
{
    make_workplan (name);
}

void ToolpathMaker::EndWorkplan()
{
    end_workplan();
}



void ToolpathMaker::Workingstep (const char * name)
{
    end_workingstep();
    make_workingstep (name);
}






// ------------------------------------------------------------
// TOOLPATH RELATED FUNCTIONS
// ------------------------------------------------------------

void ToolpathMaker::start_move(const char * label)
{
    if (!f_design) return;

    // Create a workingstep, freeform operation, feed/speed technology
    // and machine functions (coolant etc.) if needed
    //
    if (!current_op) make_workingstep(label);
    if (!current_tp) make_toolpath(label);
   
}

void ToolpathMaker::end_move()
{
    if (!f_design || !current_tp) return;

    // close out the toolpaths and write the curves, associate the
    // basis curve with the toolpath. 

    char nm[100];
    sprintf (nm, "basic curve for WS %u TP %u", ws_count, tp_count);

    // LOCATION CURVE -- There should always be a location curve.
    stp_representation_item * repitem = make_curve (nm, loc);
    if (repitem) {
	stp_representation * rep = pnewIn(f_design) stp_representation;
	rep-> name ("");
	rep-> context_of_items(make_geometry_context());
	rep-> items()-> add (repitem);

	make_action_property ("basic curve", current_tp, rep);
    }

    // AXIS CURVE -- Only for 5-axis work
    sprintf (nm, "axis curve for WS %u TP %u", ws_count, tp_count);
    repitem = make_curve (nm, axis);
    if (repitem) {
	stp_representation * rep = pnewIn(f_design) stp_representation;
	rep-> name ("");
	rep-> context_of_items(make_geometry_context());
	rep-> items()-> add (repitem);

	make_action_property ("tool axis", current_tp, rep);
    }

    current_tp = 0;
}

stp_machining_toolpath * ToolpathMaker::make_toolpath (const char * label)
{
    if (!f_design) return 0;

    // Finish up any previous toolpath
    end_move();

    RoseStringObject nm;
    nm.resize (strlen(label? label: "") + 30);
    if (!label || !*label)
	sprintf (nm, "WS %u TP %u", ws_count, ++tp_count);
    else 
	sprintf (nm, "%s WS %u TP %u", label, ws_count, ++tp_count);


    // Make a cutter location toolpath.
    stp_machining_toolpath * tp = pnewIn(f_design) stp_machining_toolpath;
    tp-> name (nm);
    tp-> description ("cutter location trajectory");
    tp-> consequence ("");
    tp-> purpose ("");

    // Associate the toolpath with the current operation using a
    // sequence relationship object
    //
    add_to_workingstep(tp);


    // Always give the toolpath a technology
    stp_machining_technology * mt = make_technology();
    stp_machining_technology_relationship * mt_rel = 
	pnewIn(f_design) stp_machining_technology_relationship;

    mt_rel-> name("");
    mt_rel-> description("");
    mt_rel-> relating_method (tp);
    mt_rel-> related_method (mt);


    // Only give the toolpath a machine functions if the coolant state
    // has changed  DO LATER


    // Rapid is handled by an overriding speed profile parameter on
    // the toolpath rather than by anything in the technology
    //
    if (is_rapid) {
	const char * shortcut = "__CONSTANT STRING-REP rapid";
	stp_representation * rapid_rep;

	// Use shortcut to rapid-rep if we have already created one
	RoseObject * obj = f_design-> findObject (shortcut);
	if (obj) rapid_rep = ROSE_CAST (stp_representation, obj);
	else {
	    // these are always shared, so just give the rep item and
	    // representation names of 'constant' to emphasise the point
	    //
	    stp_descriptive_representation_item * it = 
		pnewIn(f_design) stp_descriptive_representation_item;

	    it-> name("");
	    it-> description ("rapid");

	    rapid_rep = pnewIn(f_design) 
		stp_machining_toolpath_speed_profile_representation;
	    rapid_rep-> name ("");
	    rapid_rep-> context_of_items (make_simple_context());
	    rapid_rep-> items()-> add (it);
	}


	stp_action_property * ap = pnewIn(f_design) stp_action_property;
	ap-> name ("speed profile");
	ap-> description ("rapid");
	ap-> definition (pnewIn(f_design) stp_characterized_action_definition);
	ap-> definition()-> _action_method (tp);

	stp_action_property_representation * apr = 
	    pnewIn(f_design) stp_action_property_representation;

	apr-> name ("");
	apr-> description ("rapid");
	apr-> property (ap);
	apr-> representation (rapid_rep);
    }

    // Some common properties for the toolpaths
    //
    make_action_property (
	"trajectory type", tp, make_string_rep ("trajectory path")
	);
    
    make_action_property (
	"priority", tp, make_string_rep ("required")
	);


    // The basis, tool axis, and surface normal curves can not be
    // associated with the toolpath until they are finished writing.

    current_tp = tp;
    return tp;
}




// ------------------------------------------------------------
// WORKPLAN RELATED FUNCTIONS
// ------------------------------------------------------------

void ToolpathMaker::make_workplan (const char * label)
{
    if (!f_design) return;

    stp_machining_workplan * wp = pnewIn(f_design) stp_machining_workplan;
    wp-> name (label? label: "");
    wp-> description ("");
    wp-> consequence ("");
    wp-> purpose ("");

    // Associate the new workplan with the current workplan if it
    // exists, or as the main workplan of the project otherwise.
    //
    if (!plan_ids.size()) {
	// first one, add to the project
	stp_product_definition_process * pdp = pnewIn(f_design)
	    stp_product_definition_process;
	pdp-> name ("machining");
	pdp-> description ("");
	pdp-> identification ("");
	pdp-> chosen_method (wp);

	stp_process_product_association * ppa = pnewIn(f_design)
	    stp_process_product_association;
	ppa-> name ("");
	ppa-> description ("");
	ppa-> process (pdp);
	ppa-> defined_product (
	    pnewIn(f_design) stp_characterized_product_definition
	    );
    
	ppa-> defined_product()-> _product_definition(current_project_pd);
    }
    else {
	// add to existing workplan
	add_to_workplan (wp);
    }

    // This is now the active workplan
    plan_ids.push(wp);
    plan_idxs.push((void*)1);  // start numbering at zero

    current_op = 0;
    current_tp = 0;
}

void ToolpathMaker::end_workplan()
{
    if (!f_design || !plan_ids.size()) return;

    end_workingstep();

    // Pop back to the last workplan
    plan_ids.pop();
    plan_idxs.pop(); 

    // if we end the last workplan, close the file
    if (!plan_ids.size()) Done();
}

void ToolpathMaker::add_to_workplan(
    stp_machining_process_executable * mpe 
    )
{
    // Associate the id with the current workplan using a sequence
    // relationship object
    //
    if (!f_design) return;

    unsigned seq = (unsigned) plan_idxs.top();
    plan_idxs.pop();
    plan_idxs.push((void*) (seq+1));

    stp_machining_process_sequence_relationship * rel = pnewIn (f_design) 
	stp_machining_process_sequence_relationship;

    rel-> name("");
    rel-> description ("");
    rel-> relating_method ((stp_machining_workplan*) plan_ids.top());
    rel-> related_method (mpe);
    rel-> sequence_position (seq);
}




// ------------------------------------------------------------
// WORKINGSTEP RELATED FUNCTIONS
// ------------------------------------------------------------

// Make a workingstep and associated operation
//
void ToolpathMaker::make_workingstep (const char * label)
{
    if (!f_design) return;

    RoseStringObject nm;
    nm.resize(strlen(label? label: "") + 30);

    if (!label || !*label)
	sprintf (nm, "WS %u", ++ws_count);
    else 
	sprintf (nm, "%s WS %u", label, ++ws_count);


    stp_machining_workingstep * ws = pnewIn(f_design)
	stp_machining_workingstep;

    ws-> name (nm);
    ws-> description ("machining");
    ws-> consequence ("");
    ws-> purpose ("");

    // Associate the workingstep with the current workplan using a
    // sequence relationship object
    //
    add_to_workplan (ws);

    // Associate the workingstep with our single toolpath feature
    //
    stp_machining_feature_relationship * feat_rel = pnewIn (f_design)
	stp_machining_feature_relationship;

    feat_rel-> name("");
    feat_rel-> description("");
    feat_rel-> relating_method(ws);
    feat_rel-> related_method (make_toolpath_feature());

    // Create the operation.  This is what the toolpaths, technology,
    // and machine functions (coolant and such) are associated with so
    // that is what we return.
    //
    current_tp = 0;
    current_tpidx = 0;

    current_op = pnewIn(f_design) stp_freeform_milling_operation;
    current_op-> name (nm);
    current_op-> description("");
    current_op-> consequence("");
    current_op-> purpose("");


    // Link the operation to the workingstep
    stp_machining_operation_relationship * op_rel = pnewIn(f_design)
	stp_machining_operation_relationship;
    op_rel-> name("");
    op_rel-> description("");
    op_rel-> relating_method(ws);
    op_rel-> related_method (current_op);


    stp_machining_technology_relationship * tech_rel = pnewIn(f_design)
	stp_machining_technology_relationship;
    tech_rel-> name("");
    tech_rel-> description("");
    tech_rel-> relating_method(current_op);
    tech_rel-> related_method (make_technology());

    stp_machining_functions_relationship * func_rel = pnewIn(f_design)
	stp_machining_functions_relationship;
    func_rel-> name("");
    func_rel-> description("");
    func_rel-> relating_method(current_op);
    func_rel-> related_method (make_mfunc());

    // The workingstep operation needs a tool, use the current one or
    // create a default one if needed.
    //
    if (!current_tool) 
	LoadTool(1);

    current_tool-> users.append ((stp_machining_operation*) current_op);
}

void ToolpathMaker::end_workingstep()
{
    end_move();
    current_op = 0;
    current_tp = 0;
}


void ToolpathMaker::add_to_workingstep(
    stp_machining_toolpath * tp
    )
{
    // Associate the toolpath with the current operation using a
    // sequence relationship object
    //
    if (!f_design) return;

    stp_machining_toolpath_sequence_relationship * rel = pnewIn (f_design) 
	stp_machining_toolpath_sequence_relationship;

    rel-> name("");
    rel-> description ("");
    rel-> relating_method (current_op);
    rel-> related_method (tp);
    rel-> sequence_position (++current_tpidx);
}




stp_machining_feature_process * ToolpathMaker::make_toolpath_feature()
{
    if (!f_design) return 0;
    if (current_feature_mfp) return current_feature_mfp;


    // Only one toolpath feature is created and it is associated with
    // all of the workingsteps.  It is a step shape aspect and refers
    // to the workpiece product definition shape.
    //
    if (!current_workpiece_pds) {
	PartNo ("unspecified part");
    }

    stp_instanced_feature * feat = pnewIn(f_design) stp_instanced_feature;
    feat-> stp_characterized_object::name("");
    feat-> stp_characterized_object::description("toolpath");
    feat-> stp_shape_aspect::name("");
    feat-> stp_shape_aspect::description("toolpath");
    feat-> of_shape (current_workpiece_pds);
    feat-> product_definitional (ROSE_TRUE);


    // The machining feature process is what all of the workingsteps
    // refer to when they attach to a feature.
    //
    stp_machining_feature_process * mfp = 
	pnewIn(f_design) stp_machining_feature_process;

    mfp-> name("");
    mfp-> description("machining");
    mfp-> consequence("");
    mfp-> purpose("");

    stp_property_process * pp = pnewIn(f_design) stp_property_process;
    pp-> name("");
    pp-> description("machining");
    pp-> identification("");
    pp-> chosen_method (mfp);

    stp_process_property_association * ppa = pnewIn(f_design) 
	stp_process_property_association;

    ppa-> name("");
    ppa-> description("machining");
    ppa-> process (pp);
    ppa-> property_or_shape (pnewIn(f_design) stp_property_or_shape_select);
    rose_put_nested_object (ppa-> property_or_shape(), feat);


    // Give the feature a simple representation with a placement
    // at (0,0,0)
    //
    stp_axis2_placement_3d * ap = pnewIn(f_design) stp_axis2_placement_3d;
    ap-> name ("orientation");
    ap-> location (make_point ("origin", 0,0,0));


    stp_representation * rep = pnewIn(f_design) 
	stp_shape_representation_with_parameters;
    rep-> name("");
    rep-> context_of_items (make_geometry_context());
    rep-> items()-> add(ap);


    stp_product_definition_shape * pds = pnewIn(f_design) 
	stp_product_definition_shape;
    pds-> name ("orientation");
    pds-> description ("toolpath");
    pds-> definition (pnewIn(f_design) stp_characterized_definition);
    rose_put_nested_object (pds->definition(), feat);
    

    stp_shape_definition_representation * sdr = pnewIn(f_design) 
	stp_shape_definition_representation;
    sdr-> used_representation (rep);
    sdr-> definition (pnewIn(f_design) stp_represented_definition);
    sdr-> definition()-> _property_definition (pds);

    current_feature_mfp = mfp;
    return mfp;
}



// ------------------------------------------------------------
// GEOMETRY RELATED FUNCTIONS
// ------------------------------------------------------------


stp_cartesian_point * ToolpathMaker::make_point (
    const char * label, 
    double x, 
    double y, 
    double z)
{
    if (!f_design) return 0;

    stp_cartesian_point * pt = pnewIn(f_design) stp_cartesian_point;    

    pt-> name (label);
    pt-> coordinates()-> add(x);
    pt-> coordinates()-> add(y);
    pt-> coordinates()-> add(z);
    return pt;
}



stp_curve * ToolpathMaker::make_curve (
    const char * label, 
    curvedef &def
    )
{
    // close out an open composite curve.
    //
    stp_curve * id = 0;
    if (def.curves.size()) {

	// close out any last polyline segment
	if (def.points.size()) {
	    def.curves.append ((stp_curve*) make_polyline ("", def.points));
	    def.points.empty();
	}

	id = make_composite_curve (label, def.curves);
	def.curves.empty();
    }

    // entire basis curve is only a polyline
    else if (def.points.size()) {
	id = make_polyline (label, def.points);
	def.points.empty();
    }
    return id;
}



stp_polyline * ToolpathMaker::make_polyline (
    const char * label, 
    rose_vector &pts
    )
{
    if (!f_design || !pts.size()) return 0;

    unsigned i;
    stp_polyline * pl = pnewIn(f_design) stp_polyline;
    pl-> name (label);

    for (i=0; i< pts.size(); i++) {
	pl-> points()-> add((stp_cartesian_point*)pts[i]);
    }
    return pl;
}


stp_composite_curve * ToolpathMaker::make_composite_curve (
    const char * label, 
    rose_vector &curves
    )
{
    if (!f_design || !curves.size()) return 0;

    // These are all going to be sequential so just keep track of the
    // base id.  If we had objects in between, we would need to keep a
    // separate list.
    //

    unsigned i;
    
    stp_composite_curve * cc = pnewIn(f_design) stp_composite_curve;
    cc-> name (label? label: "");
    cc-> self_intersect (ROSE_UNKNOWN);

    for (i=0; i< curves.size(); i++) {
	stp_composite_curve_segment * seg = pnewIn(f_design) 
	    stp_composite_curve_segment;
	seg-> parent_curve ((stp_curve*)curves[i]);
	seg-> same_sense (ROSE_TRUE);
	seg-> transition (
	    (i+1 == curves.size()?
	     stp_transition_code_discontinuous: 
	     stp_transition_code_continuous)
	    );
	cc-> segments()-> add(seg);
    }

    return cc;
}

